#
/*
 */

/*
 *	DH-11 driver
 *	This driver calls on the DHDM driver.
 *	If the DH has no DM11-BB, then the latter will
 *	be fake. To insure loading of the correct DM code,
 *	lib2 should have dhdm.o, dh.o and dhfdm.o in that order.
 */

#include "../param.h"
#include "../conf.h"
#include "../user.h"
#include "../tty.h"
#include "../proc.h"

#define	DHADDR	0160020
#define	NDH11	16	/* number of lines */
#define	DHNCH	16	/* max number of DMA chars */

struct	tty dh11[NDH11];
/*
 * Place from which to do DMA on output
 */
char	dh_clist[NDH11][DHNCH];

/*
 * Used to communicate the number of lines to the DM
 */
int	ndh11	NDH11;

/*
 * Hardware control bits
 */
#define	BITS6	01
#define	BITS7	02
#define	BITS8	03
#define	TWOSB	04
#define	PENABLE	020
/* DEC manuals incorrectly say this bit causes generation of even parity. */
#define	OPAR	040
#define	HDUPLX	040000

#define	IENABLE	030100
#define	PERROR	010000
#define	FRERROR	020000
#define	XINT	0100000
#define	SSPEED	7	/* standard speed: 300 baud */

/*
 * Software copy of last dhbar
 */
int	dhsar;

struct dhregs {
	int dhcsr;
	int dhnxch;
	int dhlpr;
	int dhcar;
	int dhbcr;
	int dhbar;
	int dhbreak;
	int dhsilo;
};

/*
 * Open a DH11 line.
 */
dhopen(dev, flag)
{
	register struct tty *tp;
	extern dhstart();

	if (dev.d_minor >= NDH11) {
		u.u_error = ENXIO;
		return;
	}
	tp = &dh11[dev.d_minor];
	tp->t_addr = dhstart;
	tp->t_dev = dev;
	DHADDR->dhcsr =| IENABLE;
	tp->t_state =| WOPEN|SSTART;
	if ((tp->t_state&ISOPEN) == 0) {
		if (tp->t_speeds == 0) {
			tp->t_speeds = SSPEED | (SSPEED<<8);
			tp->t_flags = ODDP|EVENP|ECHO;
		}
		dhparam(tp);
	}
	dmopen(dev);
	tp->t_state =& ~WOPEN;
	tp->t_state =| ISOPEN;
	if (u.u_procp->p_ttyp == 0)
		u.u_procp->p_ttyp = tp;
}

/*
 * Close a DH11 line.
 */
dhclose(dev)
{
	register struct tty *tp;

	tp = &dh11[dev.d_minor];
	dmclose(dev);
	tp->t_state =& (CARR_ON|SSTART);
	wflushtty(tp);
}

/*
 * Read from a DH11 line.
 */
dhread(dev)
{
	ttread(&dh11[dev.d_minor]);
}

/*
 * write on a DH11 line
 */
dhwrite(dev)
{
	ttwrite(&dh11[dev.d_minor]);
}

/*
 * DH11 receiver interrupt.
 */
dhrint()
{
	register struct tty *tp;
	register int c;

	while ((c = DHADDR->dhnxch) < 0) {	/* char. present */
		tp = &dh11[(c>>8)&017];
		if (tp >= &dh11[NDH11])
			continue;
		if(c&FRERROR)
			c = -1;		/* break */
		ttyinput(c, tp);
	}
}

/*
 * stty/gtty for DH11
 */
dhsgtty(dev, av)
int *av;
{
	register struct tty *tp;
	register r;

	tp = &dh11[dev.d_minor];
	if (ttystty(tp, av))
		return;
	dhparam(tp);
}

/*
 * Set parameters from open or stty into the DH hardware
 * registers.
 */
dhparam(atp)
struct tty *atp;
{
	register struct tty *tp;
	register int lpr;

	tp = atp;
	spl5();
	DHADDR->dhcsr.lobyte = tp->t_dev.d_minor | IENABLE;
	/*
	 * Hang up line?
	 */
	if (tp->t_speeds.lobyte==0) {
		dmclose(tp->t_dev);
		return;
	}
	lpr = (tp->t_speeds.hibyte<<10) | (tp->t_speeds.lobyte<<6);
	if (tp->t_speeds.lobyte == 4)		/* 134.5 baud */
		lpr =| BITS7|HDUPLX; else
		if (tp->t_flags&EVENP)
			if (tp->t_flags&ODDP)
				lpr =| BITS8; else
				lpr =| BITS7|PENABLE; else
			lpr =| BITS7|OPAR|PENABLE;
	if (tp->t_speeds.lobyte == 3)	/* 110 baud */
		lpr =| TWOSB;
	DHADDR->dhlpr = lpr;
	spl0();
}

/*
 * DH11 transmitter interrupt.
 * Restart each line which used to be active but has
 * terminated transmission since the last interrupt.
 */
dhxint()
{
	register struct tty *tp;
	register ttybit, bar;

	bar = dhsar & ~DHADDR->dhbar;
	DHADDR->dhcsr =& ~XINT;
	ttybit = 1;
	for (tp = dh11; bar; tp++) {
		if(bar&ttybit) {
			dhsar =& ~ttybit;
			bar =& ~ttybit;
			tp->t_state =& ~BUSY;
			dhstart(tp);
		}
		ttybit =<< 1;
	}
}

/*
 * Start (restart) transmission on the given DH11 line.
 */
dhstart(atp)
struct tty *atp;
{
	extern ttrstrt();
	register c;
	register char *rp;
	register struct tty *tp;
	int sps, i, n;

	sps = PS->integ;
	spl5();
	tp = atp;
	if ((tp->t_state&(TIMEOUT|BUSY|HALTOP))!=0 || (n=tp->t_outq.c_cc)==0)
		goto out;
	if ((c = tp->t_nrem) == 0) {
		rp = dh_clist[tp->t_dev.d_minor];
		c = DHNCH;
		n =- TTLOWAT;
		i = c;
		do {
		if ((*rp++ = getc(&tp->t_outq)) < 0) {
				if (c==i) {
					c = *--rp;
					c =& 0177;
					if (n==0)
						wakeup(&tp->t_outq);
					goto out0;
				}
				c--;
				goto out1;
			}
		} while (--c);
out1:
		i = (c =- i);
		if(tp->t_outq.c_cc==0 ||
		   ((c =+ n) <= 0 && c > i))
			wakeup(&tp->t_outq);
		if ((c = *--rp) < 0) {
			i++;
			c++;
			tp->t_nrem = (c =& 0177);
		} else
			tp->t_nrem = 0;
	} else
		if(c > 0) {
out0:
			timeout(ttrstrt, tp, c );
			tp->t_state =| TIMEOUT;
			tp->t_nrem = 0;
			goto out;
		}
	c = tp-dh11;
	if(tp->t_nrem < 0)
		tp->t_nrem = ~tp->t_nrem;
	else {
		DHADDR->dhcsr.lobyte = c | IENABLE;
		DHADDR->dhcar = dh_clist[tp->t_dev.d_minor];
		DHADDR->dhbcr = i;
	}
	c = 1<<c;
	DHADDR->dhbar =| c;
	dhsar =| c;
	tp->t_state =| BUSY;
out:
	PS->integ = sps;
}

ttstop(atp)
{
	register struct tty *tp;
	register i;

	tp = atp;
	if((!(&dh11[0] <= tp && tp < &dh11[NDH11])) || (tp->t_state&BUSY)==0)
		return;
	tp->t_state =& ~BUSY;
	i = 1 << (tp - dh11);
	dhsar =& ~i;
	DHADDR->dhbar =& ~i;
	tp->t_nrem = ~tp->t_nrem;
}
